/*
 * midiplay : MIDI player for UNIX
 *
 * 19th, Dec. 1994. by Takanari HAYAMA
 *
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include "midi.h"

#ifdef __hpux
#  include <time.h>
#endif 

#ifdef sun
#  include <sys/time.h>
#endif

#ifndef MIDIPORT
#  ifdef __hpux
#    define MIDIPORT "/dev/tty00"
#  else
#    define MIDIPORT "/dev/ttya"
#  endif
#endif

#define BUFFER_SIZE 1024		/* 31250bps x 8s / 8bit */

#define R_NORM 1
#define R_GM   2
#define R_GS   3
#define DEFAULT_BPS 38400
#define B31250 B50

extern fdebug();

/* external variables */
extern struct midi **eventseq;
extern struct midihead header;
extern int *eventsz;
extern int *tflag;
int visual;

/* global variables */
char *com, *outdev;
static unsigned char *cptr, buffer[BUFFER_SIZE];
int std_out, std_in, bps, bcnt, sigflag;
int mididev;
int tpt, reset_mode;
int debug;
FILE *txtout;

#define GS_RESET_SZ 12
char gs_reset[] = { 0xf0, 0x41, 0x10, 0x42, 0x12, 0x40, 0x00, 
		   0x7f, 0x00, 0x41, 0xf7};

#define GM_RESET_SZ 6
char gm_reset[] = { 0xf0, 0x7e, 0x7f, 0x09, 0x01, 0xf7 };

#define ALL_RESET_SZ 96
char all_reset[] = { 0xb0, 0x78, 0x00, 0xb0, 0x79, 0x00,
		    0xb1, 0x78, 0x00, 0xb1, 0x79, 0x00,
		    0xb2, 0x78, 0x00, 0xb2, 0x79, 0x00,
		    0xb3, 0x78, 0x00, 0xb3, 0x79, 0x00,
		    0xb4, 0x78, 0x00, 0xb4, 0x79, 0x00,
		    0xb5, 0x78, 0x00, 0xb5, 0x79, 0x00,
		    0xb6, 0x78, 0x00, 0xb6, 0x79, 0x00,
		    0xb7, 0x78, 0x00, 0xb7, 0x79, 0x00,
		    0xb8, 0x78, 0x00, 0xb8, 0x79, 0x00,
		    0xb9, 0x78, 0x00, 0xb9, 0x79, 0x00,
		    0xba, 0x78, 0x00, 0xba, 0x79, 0x00,
		    0xbb, 0x78, 0x00, 0xbb, 0x79, 0x00,
		    0xbc, 0x78, 0x00, 0xbc, 0x79, 0x00,
		    0xbd, 0x78, 0x00, 0xbd, 0x79, 0x00,
		    0xbe, 0x78, 0x00, 0xbe, 0x79, 0x00,
		    0xbf, 0x78, 0x00, 0xbf, 0x79, 0x00 };

#define WRBFSIZE	16384
unsigned char wrbuff[WRBFSIZE];
int nwr;

/*
 * sigexit
 */

void sigexit()
{
  void send_reset();

  send_reset();
  exit(0);
}

/*
 * sighandler
 */

int sighandle()
{
  static fd_set x;
  static struct timeval tp = {0, 0};

  FD_ZERO(&x);
  FD_SET(0, &x);
  select(1, &x, (fd_set*)NULL, (fd_set*)NULL, &tp);
  sigflag = FD_ISSET(0, &x);
  return(1);
}

/*
 * signal setup
 */

void mysigset()
{
  signal(SIGIO, sighandle);
  signal(SIGHUP , sigexit);
  signal(SIGINT , sigexit);
  signal(SIGQUIT, sigexit);
  signal(SIGKILL, sigexit);
  signal(SIGTERM, sigexit);
}
/*
 * flush the buffer
 */

long extratime;

long myflush()
{
  struct timeval st, et;
  static long ret;

  gettimeofday(&st, NULL);
  write(mididev, wrbuff, nwr);
  nwr = 0;
  gettimeofday(&et, NULL);

  ret = (et.tv_sec - st.tv_sec) * 1000000 + (et.tv_usec - st.tv_usec)
    + extratime;
  extratime = 0;

  return(ret);
}

/*
 * write the code to the buffer
 */

void mywrite(d, n)
     unsigned char *d;
     int n;
{
  if ((nwr + n) >= WRBFSIZE)
    extratime = myflush();
  while(n--)
    wrbuff[nwr++] = *(d++);

  return;
}

/*
 * reset the MIDI inst.
 */

void send_reset()
{
  usleep(600000);
  mywrite(all_reset, ALL_RESET_SZ);
  usleep(600000);

  switch(reset_mode) {
  case R_NORM:
    break;
  case R_GM:
    mywrite(gm_reset, GM_RESET_SZ);
    usleep(700000);
    break;
  case R_GS:
    mywrite(gs_reset, GS_RESET_SZ);
    usleep(700000);
    break;
  default:
    fprintf(stderr, "%s: Unknown reset mode given.\n", com);
    exit(1);
  }
  return;
}

/*
 * open the device for play
 */

void opendevice()
{
  struct termios tio;
  static char *defdev = MIDIPORT;
  int speed;
  char *dev;

  if (std_out)
    mididev = 1;
  else if (outdev)
    dev = outdev;
  else
    dev = defdev;

  if (!std_out) {
    switch(bps) {
      case 38400:
	speed = B38400;
	break;
      case 31250:
	speed = B31250;
	break;
      default:
	fprintf(stderr, "%s: Bogus speed %d is given.\n", com, bps);
	exit(1);
    }
    
    if ((mididev = open(dev, O_RDWR | O_NDELAY)) != -1) {
      if (tcgetattr(mididev, &tio) < 0) {
	fprintf(stderr, "%s: Can't get attribute.\n", com);
	exit(1);
      }
      tio.c_iflag = 0;
      tio.c_oflag = 0;
      tio.c_cflag = CS8 | CREAD | CLOCAL;
      tio.c_lflag = 0;
      cfsetispeed(&tio, speed);
      cfsetospeed(&tio, speed);
      if (tcsetattr(mididev, TCSANOW, &tio) < 0) {
	fprintf(stderr, "%s: Can't set attribute.\n", com);
	exit(1);
      }
    }
    else {
      fprintf(stderr, "%s: Can't open device (%s).\n", com, dev);
      exit(1);
    }
  }
  myflush();
  return;
}

/*
 * assign the functinos for ecah midifile funcs.
 */

void initfuncs()
{
  Mf_getc		= myread;
  Mf_error		= fatal;
  Mf_header		= read_header;
  Mf_trackstart		= midi_starttrack;
  Mf_trackend		= midi_endtrack; /* fault tolerance */
  Mf_noteon		= midi_note_on;
  Mf_noteoff		= midi_note_off;
  Mf_pressure		= midi_pressure;
  Mf_parameter		= midi_control;
  Mf_pitchbend		= midi_pitchbend;
  Mf_program		= midi_program;
  Mf_chanpressure	= midi_chanpressure;
  Mf_sysex		= midi_sysex;
/*Mf_metamisc		= midi_metamisc;*/
/*Mf_sqspecific		= midi_sqspecific;*/
/*Mf_seqnum		= midi_sqnum;*/
  Mf_text		= midi_text;
  Mf_eot		= midi_eot;
  Mf_timesig		= midi_timesig;
/*Mf_smpte		= midi_smpte;*/
  Mf_tempo		= midi_tempo;
/*Mf_keysig		= midi_keysig;*/

  return;
}

/*
 * usage
 */

void usage()
{
  fprintf(txtout, "%s [-|-o <dev>] [-gs|-gm] [-d] [-r] <file> ...\n", com);
  exit(-1);
}

/*
 * midi event list -> midi sig. sequence
 */

#define BUFSIZE 256

int parse()
{
  int i, zz, *ptr, *p, *flag, n, xx, yy, blank;
  int omask;
  long stime;
  unsigned long ts, tu, ct;
  unsigned int tempo, tick, ctick, delta;
  unsigned char *x, pstat;
  char readbuf[BUFSIZE];
  struct timeval tp;
  extern int mf_ticks2usec();

  /* check if there are any blank track */
  for (i = 0, blank = 0; i < header.ntrks; ++i) {
    if (tflag[i] != 1)
      ++blank;
  }

  /* clear the pointer */
  ptr  = (int*)malloc(header.ntrks * sizeof(int));
  flag = (int*)malloc(header.ntrks * sizeof(int));
  for(i = 0; i < header.ntrks; ++i)
    flag[i] = ptr[i] = (tflag[i] == 1) ? 0 : 1;

  /* infinite loop */
  tempo = 500000;
  tick = ctick = 0;
  zz = blank;
  pstat = 0x00;
  while(1) {
    for(i = 0; i < header.ntrks; ++i) {
      if (sigflag) {
	int ignore = 0;
	omask = sigblock(sigmask(SIGIO));
	if ((n = read(0, readbuf, BUFSIZE)) < 0) {
	  fprintf(stderr, "%s: read error from stdin.\n", com);
	  exit(-1);
	}
	xx = 0;
	for (yy = 0; yy < n; ++yy) {
	  switch(readbuf[yy]) {
	  case '>':
	  case ' ':
	    ++xx;
	    break;
	  case '<':
	    --xx;
	    break;
	  case 's':
	  case 'S':
	    xx = 0;
	    break;
	  default:
	    ignore = 1;
	    break;
	  }
	}
	sigflag = 0;
	sigsetmask(omask);
	mysigset();
	if (!ignore) 
	  return(xx);
      }
      while ((!flag[i])&&(tick == eventseq[i][ptr[i]].ctime)) {
	if (ctick != eventseq[i][ptr[i]].ctime) {
	  delta = eventseq[i][ptr[i]].ctime - ctick;
	  stime = mf_ticks2usec(delta, header.division, tempo) - myflush();
	  if (stime > 50000)
	    usleep((unsigned long)stime);
	  else if (stime > 0) {
	    gettimeofday(&tp, 0);
	    ts = tp.tv_sec;
	    tu = tp.tv_usec;
	    do {
	      gettimeofday(&tp, 0);
	      ct = tp.tv_usec + (tp.tv_sec - ts) * 1000000;
	    } while((ct - tu) <= stime);
	  }
	  ctick = eventseq[i][ptr[i]].ctime;
	}
	if (eventseq[i][ptr[i]].type == 0) {
	  /* ordinary data */
	  mywrite(eventseq[i][ptr[i]].data, eventseq[i][ptr[i]].n);
	}
	else if (eventseq[i][ptr[i]].type == 1) {
	  /* text */
	  x = eventseq[i][ptr[i]].data;
	  if (*x == 0x01) 
	    fprintf(txtout, "%s\n", &x[1]);
	  else if (*x == 0x02)
	    fprintf(txtout, "Copyright: %s\n", &x[1]);
	  else if (*x == 0x03)
	    fprintf(txtout, "Track %d: %s\n", i, &x[1]);
	  else if (*x == 0x04)
	    fprintf(txtout, "Inst.(%d): %s\n", i, &x[1]);
	  else if (*x == 0x05)
	    fprintf(txtout, "> %s\n", &x[1]);
	  else if (*x == 0x06)
	    fprintf(txtout, "Mark: %s\n", &x[1]);
	  else if (*x == 0x07)
	    fprintf(txtout, "Cue : (%s)\n", &x[1]);
	  else {
	    fprintf(stderr, "%s: Unknown text event (%x).\n", com, *x);
	    fprintf(txtout, ">> %s\n", &x[1]);
	  }
	}
	else if (eventseq[i][ptr[i]].type == 2) {
	  /* tempo change */
	  p = (int*)eventseq[i][ptr[i]].data;
	  tempo = (unsigned int)*p;
	  if (debug)
	    fprintf(txtout, "Tempo -> %d\n", tempo);
	}
	else if (eventseq[i][ptr[i]].type == 3) {
	  /* sysex */
	  fdebug("Writing Sysex (%d,%d:%d bytes)", i, ptr[i], eventseq[i][ptr[i]].n);
	  mywrite(eventseq[i][ptr[i]].data, eventseq[i][ptr[i]].n);
	}
	else {
	  fprintf(stderr, "%s: Unknown type of event %x (NEVER!).\n", com,
		  eventseq[i][ptr[i]].type);
	  exit(-1);
	}
	if (++ptr[i] >= eventsz[i]) {
	  flag[i] = 1;
	  ptr[i] = 0;
	  if (++zz >= header.ntrks) {
	    myflush();
	    return(1);
	  }
	}
      }
    }
    ++tick;
  }
}

/*
 * main routine
 */

void main(argc, argv, envp)
     int argc;
     char **argv, **envp;
{
  struct timeval tp;
  char **mlist, *tmpp;
  int mn, rp, i, j;
  void mfread(), closemf(), freeinfo();

  /* init. */
  com = *(argv++);  /* my command name */
  cptr = buffer;
  std_out = std_in = debug = 0;
  reset_mode = R_NORM;
  bps = DEFAULT_BPS;
  outdev = NULL;
  txtout = stdout;
  nwr = visual = rp = 0;
  mlist = (char**)malloc(sizeof(char*) * argc);
  sigflag = 0;

  /* init signal */
  mysigset();

  /* term setting */
  {
    struct termios tiostat;
    tcgetattr(0, &tiostat);
    tiostat.c_lflag &= ~(ICANON | ECHO);
    tiostat.c_cc[VMIN]  = 1;
    tiostat.c_cc[VTIME] = 0;
    tcsetattr(0, TCSANOW, &tiostat);
  }

#ifdef __hpux
  {
    static int junk;
    junk = getpid();
    if (ioctl(0, FIOSSAIOOWN, &junk) == -1) {
      fprintf(stderr, "%s: error occured while FIOSSAIOOWN.\n", com);
      exit(-1);
    }
    junk = 1;
    if (ioctl(0, FIOSSAIOSTAT, &junk)) {
      fprintf(stderr, "%s: error occured while FIOSSAIOSTAT.\n", com);
      exit(-1);
    }
  }
#else /*!__hpux*/
#  ifdef F_SETOWN
  if (fcntl(0, F_SETOWN, getpid()) < 0) {
    fprintf(stderr, "%s: error occured while F_SETOWN.\n", com);
    exit(-1);
  }
#  endif
#  ifdef F_SETFL
#    ifdef FASYNC
  if (fcntl(0, F_SETFL, FASYNC) < 0) {
    fprintf(stderr, "%s: error occured while F_SETFL.\n", com);
    exit(-1);
  }
#    endif
#  endif
#endif

  /* scan arguments */
  while(*argv) {
    if (strcmp(*argv, "-") == 0) {
      txtout = stderr;
      std_out = 1;
    }
    else if (strcmp(*argv, "-d") == 0)
       debug = 1;
    else if (strcmp(*argv, "-r") == 0)
       rp = 1;
    else if (strcmp(*argv, "-gs") == 0)
      reset_mode = R_GS;
    else if (strcmp(*argv, "-gm") == 0)
      reset_mode = R_GM;
    else if (strcmp(*argv, "-o") == 0)
      outdev = *(++argv);
    else if (*argv[0] == '-') {
      fprintf(stderr, "%s: Unknown option '%s'.\n", com, *argv);
      usage();
    }
    else
      break;
    ++argv;
  }

  /* true if I should read the std midi data from stdin */
  if (*argv == NULL)
    std_in = 1;

  /* open the device */
  opendevice();

  /* initialize the functions for midifile */
  initfuncs();

  if (std_in) {
    /* open the std midi file */
    openmf(NULL);
    /* Ok. start parsing std. midi file */
    mfread();
    closemf();
    /* put the reset code first */
    send_reset();
    /* now create the MIDI signal sequence */
    parse();
    /* release memory */
    freeinfo();
  }
  else {
    mn = 0;
    while(argv[mn])
      ++mn;
    for(i = 0; i < mn; ++i)
      mlist[i] = argv[i];
    if (rp) {
      gettimeofday(&tp, NULL);
      srand(tp.tv_usec);
      for(i = 0; i < mn; ++i) {
	j = rand() % mn;
	tmpp = mlist[i];
	mlist[i] = mlist[j];
	mlist[j] = tmpp;
      }
    }
    for(i = 0; i < mn;) {
      printf("MUSIC: %s\n", mlist[i]);
      /* open the std midi file */
      openmf(mlist[i]);
      /* Ok. start parsing std. midi file */
      mfread();
      closemf();
      /* put the reset code first */
      send_reset();
      /* now create the MIDI signal sequence */
      i += parse();
      if (i < 0)
	i = 0;
      /* release memory */
      freeinfo();
      usleep(100000);
    }
  }

  /* better to reset before we close */
  send_reset();

  close(mididev);
  return;
}
